/*
FILE:     CifSchemaMap.h
*/
/*
VERSION:  4.200
*/
/*
DATE:     11/18/2013
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2013 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               RCSB PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/


/*!
** \file CifSchemaMap.h
**
** \brief Header file for DbLoader related classes.
*/


#ifndef CIFSCHEMAMAP_H
#define CIFSCHEMAMAP_H


#include <string>
#include <ostream>
#include <fstream>

#include "CifFile.h"
#include "SchemaMap.h"
#include "Db.h"
#include "DbOutput.h"



/**
**  \class DbOracle
**
**  \brief Oracle database class.
**
**  This class represents an Oracle database.
*/
class DbOracle : public Db
{

  // Oracle 8.1.6 SQLLDR

  public:
    DbOracle(SchemaMap& schemaMapping,
      const string& dbName = DB_DEFAULT_NAME);
    ~DbOracle();

    void WriteSchemaStart(std::ostream& io);

    void WriteLoadingStart(std::ostream& io);
    void WriteLoadingTable(std::ostream& io, const string& tableName,
      const string& path);

    void GetDate(string& dType);
    void GetText(string& dType, const unsigned int width);

    void WriteNull(std::ostream& io, const int iNull,
      const unsigned int curr, const unsigned int attSize);
    void WriteTableIndex(std::ostream& io, const string& tableNameDb,
      const vector<string>& indexList);

    void WriteNewLine(std::ostream& io, bool special = false);
};


/**
**  \class DbDb2
**
**  \brief DB2 database class.
**
**  This class represents a DB2 database.
*/
class DbDb2 : public Db
{

  public:
    DbDb2(SchemaMap& schemaMapping,
      const string& dbName = DB_DEFAULT_NAME);
    ~DbDb2();


    void GetStart(string& start);

    void WriteSchemaStart(std::ostream& io);

    void WriteLoadingStart(std::ostream& io);
    void WriteLoadingEnd(std::ostream& io);
    void WriteLoadingTable(std::ostream& io, const string& tableName,
      const string& path);

    void GetFloat(string& dType);
    void GetDate(string& dType);
    void GetText(string& dType, const unsigned int width);
    void WriteNull(std::ostream& io, const int iNull,
      const unsigned int curr, const unsigned int attSize);

    void WriteTableIndex(std::ostream& io, const string& tableNameDb,
      const vector<string>& indexList);

    void WriteBcpDoubleQuotes(std::ostream& io);
};


/**
**  \class DbMySql
**
**  \brief MySQL database class.
**
**  This class represents a MySQL database.
*/
class DbMySql : public Db
{

  public:
    DbMySql(SchemaMap& schemaMapping,
      const std::string& dbName = DB_DEFAULT_NAME,
      const std::string& dbHost = std::string());
    ~DbMySql();

    void GetStart(string& start);

    void DropTableSql(std::ostream& io, const string& tableNameDb);

    void WriteLoad(std::ostream& io);
    void WriteLoadingStart(std::ostream& io);
    void WriteLoadingTable(std::ostream& io, const string& tableName,
      const string& path);

    void WriteTableIndex(std::ostream& io, const string& tableNameDb,
      const vector<string>& indexList);

    void WriteNull(std::ostream& io, const int iNull,
      const unsigned int curr, const unsigned int attSize);

    void WriteNewLine(std::ostream& io, bool special = false);

  private:
    static const string _SQL_LOADING_FILE;

};


/**
**  \class DbSybase
**
**  \brief Sybase database class.
**
**  This class represents a Sybase database.
*/
class DbSybase : public Db
{
  public:
    DbSybase(SchemaMap& schemaMapping,
      const string& dbName = DB_DEFAULT_NAME);
    ~DbSybase();

    void GetStart(string& start);

    void WriteLoadingStart(std::ostream& io);
    void WriteLoadingTable(std::ostream& io, const string& tableName,
      const string& path);

    void WritePrint(std::ostream& io, const string& tableNameDb);
    void WriteNull(std::ostream& io, const int iNull,
      const unsigned int curr, const unsigned int attSize);

    void WriteTableIndex(std::ostream& io, const string& tableNameDb,
      const vector<string>& indexList);

    void GetDateAndTime(string& dateAndTime);

#ifdef VLAD_DATE_OBSOLETE
    void ConvertDate(string& dbDate, const string& cifDate);
#endif
};


/**
**  \class BcpOutput
**
**  \brief BCP output class.
**
**  This class represents a BCP output.  It re-implements methods
**  for data and loading scripts generation.
*/
class BcpOutput : public DbOutput
{
  public:
    BcpOutput(Db& db);
    virtual ~BcpOutput();

    void WriteDataLoadingScripts(const string& path = std::string());
    void WriteData(Block& block, const string& path = std::string());

  private:
    static const string _DATA_DELETE_FILE;

    void WriteDataLoadingScript(const string& path);
    void WriteDataLoadingFile(const string& path = std::string());

    void WriteDelete(std::ostream& io);

    void WriteEmptyString(std::ostream& io);

    void WriteSpecialDateChar(std::ostream& io, const char& specDateChar);
};
 

/**
**  \class SqlOutput
**
**  \brief SQL output class.
**
**  This class represents an SQL output. It re-implements methods
**  for schema, data and loading scripts generation.
*/
class SqlOutput : public DbOutput
{

  public:
    SqlOutput(Db& db);
    virtual ~SqlOutput();

    void WriteSchema(const string& path = std::string());
    void WriteDataLoadingScripts(const string& path = std::string());
    void WriteData(Block& block, const string& path = std::string());

  protected:
    void WriteEmptyNumeric(std::ostream& io);
    bool IsFirstTextNewLineSpecial();
    void WriteNewLine(std::ostream& io, bool special = false);
    void GetTableStart(string& tableStart, const string& tableName);
    void GetTableEnd(string& tableEnd);

  private:
    static const unsigned int _MAX_SQL_NAME_LENGTH = 60;
    static const string _SCHEMA_LOADING_SCRIPT;
    static const string _SCHEMA_DELETE_FILE;
    static const string _DATA_FILE;

    void WriteSqlScriptSchemaInfo(std::ostream& io);
    void WriteDataLoadingScript(const string& path);

    void CreateTableSql(std::ostream& io, const string& tableName);

    void WriteAuxTables(std::ostream& io, ISTable* infoP,
      const vector<string>& tableNames);

    void WriteEmptyString(std::ostream& io);
    void WriteEmptyDate(std::ostream& io);
};


/**
**  \class DbLoader
**
**  \brief The controller of CIF to DB data conversion process.
**
**  This class controls the conversion process of CIF data to DB loadable data.
**  It provides methods for its construction/destruction, conversion of
**  ASCII CIF files, conversion of serialized (binary) CIF files, conversion of
**  in-memory CIF file objects, setting the working directory. Data conversion
**  methods can be instructed to generate: only the data loading files, only
**  the data loading shell scripts or both the data and scripts. DB server
**  data loading is accomplished by executing the generated loading shell
**  scripts, which utilize the data loading files, obtained during the
**  conversion process.
*/
class DbLoader
{
  public:
    enum eConvOpt
    {
        // Generate only data loading files
        eDATA_ONLY = 0,

        // Generate both data loading files and loading shell scripts
        eDATA_WITH_SCRIPTS,

        // Generate only loading shell scripts
        eSCRIPTS_ONLY
    };

    /**
    **  Constructs a CIF to DB controller object.
    **  
    **  \param[in] schemaMapping - reference to the schema mapping object
    **  \param[in] dbOutput - reference to the database output object
    **  \param[in] verbose - optional parameter that indicates whether
    **    logging should be turned on (if true) or off (if false).
    **    If \e verbose is not specified, logging is turned off.
    **  \param[in] workDir - optional parameter that indicates the working
    **    directory in which the files will be placed. If \e workDir is
    **    not specified, current process working directory is used.
    **
    **  \return Not applicable
    **
    **  \pre None
    **
    **  \post None
    **
    **  \exception: None
    */
    DbLoader(SchemaMap& schemaMapping, DbOutput& dbOutput,
      bool verbose = false, const string& workDir = std::string());

    /**
    **  Destructs a CIF to DB controller object.
    **
    **  \param: Not applicable
    **
    **  \return Not applicable
    **
    **  \pre None
    **
    **  \post None
    **
    **  \exception: None
    */
    virtual ~DbLoader();

    /**
    **  Sets the working directory in which the generated files will be placed.
    **
    **  \param[in] workDir - indicates the working directory in which the
    **    files will be placed.
    **
    **  \return None
    **
    **  \pre None
    **
    **  \post None
    **
    **  \exception: None
    */
    void SetWorkDir(const string& workDir);

    /**
    **  Converts an ASCII CIF file to DB loadable data with/without generating
    **  loading scripts. The specified file is first parsed and then converted.
    **
    **  \param[in] asciiFile - indicates the name of the ASCII CIF file,
    **    which is to be converted.
    **  \param[in] convOpt - indicates data conversion options: generate
    **    data only, generate data with loading scripts, generate loading
    **    scripts only.
    **
    **  \return None
    **
    **  \pre None
    **
    **  \post None
    **
    **  \exception: None
    */
    void AsciiFileToDb(const string& asciiFile, const eConvOpt convOpt);

    /**
    **  Converts a serialized (binary) CIF file to DB loadable data
    **  with/without generating loading scripts. The specified file is
    **  de-serialized (without the need to parse it) and converted.
    **  
    **  \param[in] serFile - indicates the name of the serialized (binary)
    **    CIF file, which is to be converted.
    **  \param[in] convOpt - indicates data conversion options: generate
    **    data only, generate data with loading scripts, generate loading
    **    scripts only.
    **
    **  \return None
    **
    **  \pre None
    **
    **  \post None
    **
    **  \exception: None
    */
    void SerFileToDb(const string& serFile, const eConvOpt convOpt);

    /**
    **  Converts an in-memory CIF file object to DB loadable data
    **  with/without generating loading scripts.
    **  
    **  \param[in] cifFile - a reference to the in-memory CIF file object,
    **    which is to be converted.
    **  \param[in] convOpt - indicates data conversion options: generate
    **    data only, generate data with loading scripts, generate loading
    **    scripts only.
    **
    **  \return None
    **
    **  \pre None
    **
    **  \post None
    **
    **  \exception: None
    */
    void FileObjToDb(CifFile& cifFile, const eConvOpt convOpt);

#ifdef DB_HASH_ID
    void SetHashMode(int mode);
#endif

  private:
    static const string _LOG_FILE;

    string _workDir; // Working directory for all generated files.
    string _INPUT_FILE;

#ifdef DB_HASH_ID
    string _HASH_ID;
    int _hashMode;
#endif

    // Block name of loadable data in mmCIF format
    string _blockName;

    bool _verbose;

    std::ofstream _log;

    SchemaMap& _schemaMapping;
    DbOutput& _dbOutput;

    void _LoadBlock(Block& rBlock, Block& wBlock);

    bool _Search(vector<vector<string> >& dMap, const unsigned int iAttr,
      ISTable* isTableP, const string& blockName,
      const vector<string>& cNameMap, const string& sItem,
      const string& sCnd, const string& sFnct);
 
    void _DoFunc(vector<string>& s, const vector<string>& r,
      const string& sFnct);

    void _OpenLog(const string& logName);

    int _GetMapColumnIndex(const vector<string>& cNameMap, const string& vOf);
    void _GetMapColumnValue(string& p, vector<string>& dMapVec,
      unsigned int irow);

    void CreateTables(CifFile& writeCifFile, CifFile& readCifFile);

#ifdef VLAD_DATE_OBSOLETE
    void _dformat_1(const char *date, char *odate);
    void _dformat_2(const char *date, char *odate);
    void _dformat_3(const char *date, char *odate, int shortFlag);
    void _dformat_4(const char *date, char *odate);
    void _dformat_5(const char *date, char *odate);
#endif

    void _ReorderName(string& res, char *string, int mode);
    void _ToUpperString(string& aString);
    void _StripString(string& aString, int mode);

#ifdef DB_HASH_ID
    long long pdbIdHash(const string& id);
#endif

    static void CleanString(string& aString);

    void Clear();
};


#endif
