/*
FILE:     Db.h
*/
/*
VERSION:  4.200
*/
/*
DATE:     11/18/2013
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2013 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               RCSB PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/


/*!
** \file Db.h
**
** \brief Header file for Db class.
*/


#ifndef DB_H
#define DB_H


#include <string>
#include <vector>
#include <ostream>

#include "SchemaMap.h"


/**
**  \class Db
**
**  \brief Base class for all databases.
**
**  This class encapsulates general characteristics of each database. It
**  contains methods for generic database related operations and virtual
**  methods, that are to be specialized in concrete databases.
*/
class Db
{
  public:
    static const std::string DB_DEFAULT_NAME;

    SchemaMap& _schemaMapping;

    Db(SchemaMap& schemaMapping, const std::string& dbName = DB_DEFAULT_NAME);
    virtual ~Db();

    void SetUseOnlyPopulated(bool mode=true);
    bool GetUseOnlyPopulated();

    void SetAppendFlag(const bool appendFlag);
    bool GetAppendFlag();

    void SetFieldSeparator(const std::string& fieldSeparator);
    void SetRowSeparator(const std::string& rowSeparator);

    std::string GetCommandTerm();

    std::string GetFieldSeparator();
    std::string GetRowSeparator();

    virtual void GetStart(std::string& start);

    virtual void WriteSchemaStart(std::ostream& io);
    void WriteDeleteTable(std::ostream& io, const std::string& table,
      const std::string& where, const std::string& what);

    virtual void DropTableSql(std::ostream& io, const std::string& tableNameDb);

    virtual const std::string& GetExec();
    virtual const std::string& GetExecOption();
    virtual const std::string& GetUserOption();
    virtual const std::string& GetPassOption();

    virtual const std::string& GetEnvDbUser();
    virtual const std::string& GetEnvDbPass();

    virtual const std::string& GetConnect();
    virtual const std::string& GetTerminate();
    virtual const std::string& GetDbCommand();

    virtual void WriteLoad(std::ostream& io);

    const std::string& GetDataLoadingFileName();

    virtual void WriteLoadingStart(std::ostream& io);
    virtual void WriteLoadingEnd(std::ostream& io);
    virtual void WriteLoadingTable(std::ostream& io,
      const std::string& tableName, const std::string& path);

    virtual void WritePrint(std::ostream& io, const std::string& tableNameDb);

    virtual void GetChar(std::string& dType, const unsigned int width);
    virtual void GetFloat(std::string& dType);
    virtual void GetText(std::string& dType, const unsigned int width);
    virtual void GetDate(std::string& dType);
    virtual void WriteNull(std::ostream& io, const int iNull,
      const unsigned int curr, const unsigned int attSize);
    virtual void WriteTableIndex(std::ostream& io,
      const std::string& tableNameDb,
      const vector<std::string>& indexList);

    const std::string& GetBcpStringDelimiter();
    virtual void WriteBcpDoubleQuotes(std::ostream& io);

    virtual void WriteNewLine(std::ostream& io, bool special = false);

    bool IsFirstTextNewLineSpecial();

    virtual void ConvertDate(std::string& dbDate, const std::string& cifDate);
    virtual void ConvertTimestamp(std::string& dbTimestamp,
      const std::string& cifTimestamp);

    virtual void GetDateAndTime(std::string& dateAndTime);

  protected:
    bool _useOnlyPopulated;

    bool _appendFlag;

    // Field and row separators for compact output (eg. BCP)
    std::string _fieldSeparator; 
    std::string _rowSeparator;   

    std::string _cmdTerm;            // SQL command terminator.

    // Used in assigning permissions when exporting schema in SQL. 
    std::string _dbName; // Target database name 

    std::string _exec;
    std::string _verboseOption;
    std::string _execOption;
    std::string _hostOption;
    std::string _userOption;
    std::string _passOption;

    std::string _connect;
    std::string _terminate;
    std::string _dbCommand;
    std::string _envDbUser;
    std::string _envDbPass;

    std::string _dbHost;

    std::string _dataLoadingFileName;

    std::string _bcpStringDelimiter;

    bool _firstTextNewLineSpecial;

  private:
    static const std::string _SCRIPT_LOADING_FILE;

};

#endif
