/*
FILE:     PdbMlSchema.C
*/
/*
VERSION:  4.200
*/
/*
DATE:     11/18/2013
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2013 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               RCSB PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/


// For localtime, time, tm on SGI platform. Investigate and remove this.
#include "time.h"

#include <stdexcept>
#include <vector>
#include <set>
#include <map>
#include <sstream>
#include <iostream>
#include <algorithm>

#include "rcsb_types.h"
#include "GenString.h"
#include "CifString.h"
#include "GenCont.h"
#include "CifFile.h"
#include "CifFileUtil.h"
#include "CifExcept.h"
#include "ParentChild.h"
#include "XsdWriter.h"
#include "PdbMlWriter.h"
#include "PdbMlSchema.h"


using std::exception;
using std::pair;
using std::make_pair;
using std::string;
using std::vector;
using std::set;
using std::map;
using std::multimap;
using std::ostringstream;
using std::cerr;
using std::endl;
using std::sort;


const string PdbMlSchema::DATABLOCK_ELEMENT("datablock");
const string PdbMlSchema::DATABLOCK_NAME_ATTRIBUTE("datablockName");

const string PdbMlSchema::CATEGORY_ELEMENT_SUFFIX("Category");
const string PdbMlSchema::CATEGORY_TYPE_SUFFIX("Type");


static void format_cif_example(string& obuf, const string& str);
static void format_cif_descriptionXML(string& obuf, const string& str);
static void GetLineFromMultiline(string& line, string::size_type& iPos,
  const string& multiLine);
static void GetWordInLine(string& word, string::size_type& currLinePos,
  const string& line);


const string PdbMlSchema::SCHEMA_LOCATION_DIRECTORY
  ("http://pdbml.pdb.org/schema/");


PdbMlSchema::PdbMlSchema(XsdWriter& xsdWriter, ParentChild& parentChild,
  DataInfo& dataInfo, const string& ns, const string& prefix) :
  _xsdWriter(xsdWriter), _parentChild(parentChild), _dataInfo(dataInfo),
  _ns(ns), _prefix(prefix)
{
    if (!_ns.empty())
    {
        _nsPrefix = _ns + ":";
    }
}


void PdbMlSchema::Convert()
{
    string dictVer;
    _dataInfo.GetVersion(dictVer);

    if (dictVer.empty())
        dictVer = "1.00";

    _xsdWriter.WriteDeclaration();
    _xsdWriter.WriteNewLine();

    _WriteSchemaComment(dictVer);
    _xsdWriter.WriteNewLine();

    _xsdWriter.WriteSchemaOpeningTag();

    _WriteSchemaAttributes(_ns);
    _xsdWriter.WriteNewLine();

    vector<string> categories = _dataInfo.GetCatNames();
    sort(categories.begin(), categories.end());

    vector<string> items = _dataInfo.GetItemsNames();
    sort(items.begin(), items.end());

    _xsdWriter.IncrementIndent();

    _WriteCategoriesTypes(categories, items);

    _WriteDatablockType(categories);

    _WriteDatablockElement(categories);

    _xsdWriter.DecrementIndent();

    _xsdWriter.WriteSchemaClosingTag();
}


void PdbMlSchema::MakeSchemaFileName(string& schemaFileName,
   const string& prefix, const string& dictVer)
{
    schemaFileName = prefix;

    if (!dictVer.empty())
        schemaFileName += "-v" + dictVer;

    schemaFileName += ".xsd";
}


void PdbMlSchema::MakeFullSchemaFileName(string& schemaFileName,
   const string& prefix, const string& dictVer)
{
    schemaFileName = SCHEMA_LOCATION_DIRECTORY;

    string append;
    MakeSchemaFileName(append, prefix, dictVer);

    schemaFileName += append;
}


void PdbMlSchema::_WriteSchemaAttributes(const string& ns)
{
    _xsdWriter.WriteXsdNamespace();
    _xsdWriter.WriteNewLine();

    _xsdWriter.IncrementIndent();

    string fullSchemaFileName;
    MakeFullSchemaFileName(fullSchemaFileName, _prefix);
 
    _xsdWriter.Indent();
    _xsdWriter.WriteNamespaceAttribute(ns, fullSchemaFileName, true);
    _xsdWriter.WriteNewLine();

    _xsdWriter.Indent();
    _xsdWriter.WriteTargetNamespaceAttribute(fullSchemaFileName);
    _xsdWriter.WriteNewLine();
 
    _xsdWriter.Indent();
    _xsdWriter.WriteElementFormDefaultAttribute("qualified");
    _xsdWriter.WriteAttributeFormDefaultAttribute("unqualified");
    _xsdWriter.WriteClosingBracket();

    _xsdWriter.DecrementIndent();
}


void PdbMlSchema::_WriteSchemaComment(const string& dictVer)
{
    char timeString[31];

    time_t curr_time;
    time(&curr_time);

    struct tm* tmptr = localtime(&curr_time);
    strftime(timeString, 30, "%Y-%m-%d", tmptr);

    _xsdWriter.WriteComment("XSD type schema generated on "
      + string(timeString));

    string fullSchemaFileName;
    MakeFullSchemaFileName(fullSchemaFileName, _prefix, dictVer);
    _xsdWriter.WriteComment("Schema file: " + fullSchemaFileName);

    _xsdWriter.WriteComment("Please direct questions or comments to "\
      "John Westbrook (jwest@rcsb.rutgers.edu)");
}


void PdbMlSchema::_WriteCategoriesTypes(const vector<string>& categories,
  const vector<string>& items)
{
    for (unsigned int i=0; i < categories.size(); i++)
    {
        _xsdWriter.Indent();
        _xsdWriter.WriteComplexTypeOpeningTag();

        string catTypeName;
        PdbMlSchema::MakeCategoryTypeName(catTypeName, categories[i]);
        _xsdWriter.WriteNameAttribute(catTypeName);
        _xsdWriter.WriteClosingBracket();

        _xsdWriter.IncrementIndent();

        // Category level documentation  - 
        _WriteCategoryDocumentation(categories[i]);

        _xsdWriter.Indent();
        _xsdWriter.WriteSequenceOpeningTag();

        _xsdWriter.IncrementIndent();

        _xsdWriter.Indent();
        _xsdWriter.WriteElementOpeningTag();
        _xsdWriter.WriteNameAttribute(categories[i]);
        _xsdWriter.WriteMinOccursAttribute("0");
        _xsdWriter.WriteMaxOccursAttribute("unbounded");
        _xsdWriter.WriteClosingBracket();

        _xsdWriter.IncrementIndent();

        _xsdWriter.Indent();
        _xsdWriter.WriteComplexTypeOpeningTag(true);

        _xsdWriter.IncrementIndent();

        vector<string> keys = _dataInfo.GetCatKeys(categories[i]);
        sort(keys.begin(), keys.end());

        // Figure out the non-key category items
        vector<string> nonKeyItems;
        for (unsigned int j = 0; j < items.size(); ++j)
        {
            string categoryName;
            CifString::GetCategoryFromCifItem(categoryName, items[j]);

            if (categoryName.empty())
                continue;

            if (!String::IsCiEqual(categoryName, categories[i]))
                continue;

            if (GenCont::IsInVector(items[j], keys))
                continue;

            // Non-key item of this cateogory detected
            nonKeyItems.push_back(items[j]);
        }

        if (!nonKeyItems.empty())
        {
            _xsdWriter.Indent();
            _xsdWriter.WriteAllOpeningTag();

            _xsdWriter.IncrementIndent();
            for (unsigned int j = 0; j < nonKeyItems.size(); ++j)
            {
                _WriteNonKeyItem(nonKeyItems[j]);
            }

            _xsdWriter.DecrementIndent();

            _xsdWriter.Indent();
            _xsdWriter.WriteAllClosingTag();
        }

        for (unsigned int j = 0; j < keys.size(); ++j)
        {
            if (keys[j] == CifString::UnknownValue)
                continue;

            // Keys as attributes
            _WriteKeyItem(keys[j]);
        }

        _xsdWriter.DecrementIndent();

        _xsdWriter.Indent();
        _xsdWriter.WriteComplexTypeClosingTag();

        _xsdWriter.DecrementIndent();

        _xsdWriter.Indent();
        _xsdWriter.WriteElementClosingTag();

        _xsdWriter.DecrementIndent();

        _xsdWriter.Indent();
        _xsdWriter.WriteSequenceClosingTag();

        _xsdWriter.DecrementIndent();

        _xsdWriter.Indent();
        _xsdWriter.WriteComplexTypeClosingTag();

        _xsdWriter.WriteNewLine();
    }
}


void PdbMlSchema::_WriteDatablockType(const vector<string>& categories)
{
    // Define datablockType as unordered list of category elements!
    // Add <unique> and <key> elements here.

    _xsdWriter.Indent();
    _xsdWriter.WriteComplexTypeOpeningTag();
    string dataBlockCatTypeName;
    PdbMlSchema::MakeCategoryTypeName(dataBlockCatTypeName,
      DATABLOCK_ELEMENT);
    _xsdWriter.WriteNameAttribute(dataBlockCatTypeName);
    _xsdWriter.WriteClosingBracket();

    _xsdWriter.IncrementIndent();

    _xsdWriter.Indent();
    _xsdWriter.WriteAllOpeningTag();

    _xsdWriter.IncrementIndent();
    for (unsigned int i=0; i < categories.size(); i++)
    {
        _xsdWriter.Indent();
        _xsdWriter.WriteElementOpeningTag();

        string catElemName;
        PdbMlSchema::MakeCategoryElementName(catElemName, categories[i]);
        _xsdWriter.WriteNameAttribute(catElemName);

        string catTypeName;
        PdbMlSchema::MakeCategoryTypeName(catTypeName, categories[i]);
        _xsdWriter.WriteTypeAttribute(_nsPrefix + catTypeName);
        _xsdWriter.WriteMinOccursAttribute("0");
        _xsdWriter.WriteMaxOccursAttribute("1");
        _xsdWriter.WriteClosingBracket();

        _xsdWriter.Indent();
        _xsdWriter.WriteElementClosingTag();
    }

    _xsdWriter.DecrementIndent();

    _xsdWriter.Indent();
    _xsdWriter.WriteAllClosingTag();

    _xsdWriter.Indent();
    _xsdWriter.WriteAttributeOpeningTag();
    _xsdWriter.WriteNameAttribute(DATABLOCK_NAME_ATTRIBUTE);
    _xsdWriter._WriteDataTypeAsAttribute(eTYPE_CODE_STRING);
    _xsdWriter.WriteUseAttribute("optional");
    _xsdWriter.WriteClosingTag();

    _xsdWriter.DecrementIndent();

    _xsdWriter.Indent();
    _xsdWriter.WriteComplexTypeClosingTag();
}


void PdbMlSchema::_WriteDatablockElement(const vector<string>& categories)
{
    //  Instance of the datablock element including key/keyref elements.

    _xsdWriter.Indent();
    _xsdWriter.WriteElementOpeningTag();

    _xsdWriter.WriteNameAttribute(DATABLOCK_ELEMENT);
    string dataBlockTypeName;
    PdbMlSchema::MakeCategoryTypeName(dataBlockTypeName,
      DATABLOCK_ELEMENT);
    _xsdWriter.WriteTypeAttribute(_nsPrefix + dataBlockTypeName);
    _xsdWriter.WriteClosingBracket();

    _xsdWriter.IncrementIndent();

    for (unsigned int i=0; i < categories.size(); i++)
    {
        _WriteCategoryKeys(categories[i]);
        _WriteCategoryKeysAndKeyrefs(categories[i]); 
    }

    _xsdWriter.DecrementIndent();

    _xsdWriter.Indent();
    _xsdWriter.WriteElementClosingTag();
}


void PdbMlSchema::_WriteCategoryDocumentation(const string& catName) 
{
    // Category level documentation
    if (catName.empty())
    {
        cerr << "ERROR: Empty category name detected in \""\
          "PdbMlSchema::_WriteCategoryDocumentation()\"" << endl;

        return;
    }

    _xsdWriter.Indent();
    _xsdWriter.WriteAnnotationOpeningTag();

    _xsdWriter.IncrementIndent();

    _xsdWriter.Indent();
    _xsdWriter.WriteDocumentationOpeningTag();
    _xsdWriter.WriteLangAttribute();
    _xsdWriter.WriteClosingBracket();

    const vector<string>& categoryDescription = _dataInfo.GetCatAttribute(
      catName, CifString::CIF_DDL_CATEGORY_CATEGORY,
      CifString::CIF_DDL_ITEM_DESCRIPTION);

    string descriptionXML;
    format_cif_descriptionXML(descriptionXML, categoryDescription[0]);

    _xsdWriter._FormatTextDataXML(descriptionXML);
    if (!CifString::IsEmptyValue(descriptionXML))
        _xsdWriter.WriteNewLine();

    const vector<string>& exampleCase = _dataInfo.GetCatAttribute(catName,
      CifString::CIF_DDL_CATEGORY_CATEGORY_EXAMPLES,
      CifString::CIF_DDL_ITEM_CASE);

    const vector<string>& exampleDetail = _dataInfo.GetCatAttribute(catName,
      CifString::CIF_DDL_CATEGORY_CATEGORY_EXAMPLES,
      CifString::CIF_DDL_ITEM_DETAIL);

    for (unsigned int i = 0; i < exampleCase.size(); ++i)
    {
        _xsdWriter._FormatTextDataXML(exampleDetail[i]);
        if (!CifString::IsEmptyValue(exampleDetail[i]))
            _xsdWriter.WriteNewLine();

        if (CifString::IsEmptyValue(exampleCase[i]))
        {
            continue;
        }

        string exampleCIF;
        format_cif_example(exampleCIF, exampleCase[i]);

        try
        {
            ostringstream exampleXMLStream;

            CifFile* fobjIn = ParseCifString(exampleCIF);

            const string& parsingDiags = fobjIn->GetParsingDiags();

            if (!parsingDiags.empty())
            {
                cerr << "Diags for category " << catName << "  = " <<
                  parsingDiags << endl;
            }

            PdbMlWriter pdbMlWriter(exampleXMLStream, _ns, _dataInfo);

            vector<string> blockNames;
            fobjIn->GetBlockNames(blockNames);
            for (unsigned int ib = 0; ib < blockNames.size(); ++ib)
            {
                const string& blockName = blockNames[ib];
                Block& block = fobjIn->GetBlock(blockName);
                vector<string> tableNames;
                block.GetTableNames(tableNames);
                for (unsigned int it = 0; it < tableNames.size(); ++it)
                {
                    if (!_dataInfo.IsCatDefined(tableNames[it]))
                    {
                        cerr << " Skipping conversion to XML of the unknown "\
                          "table \"" << tableNames[it] << "\"" << endl;
                        continue;
                    }

                    ISTable* t = block.GetTablePtr(tableNames[it]);
                    if ((t != NULL) && (t->GetNumRows() > 0))
                    {
                        vector<unsigned int> emptyWidths;
                        pdbMlWriter.WriteTable(t, emptyWidths);
                    }
                }
            }

            delete (fobjIn);

            _xsdWriter._FormatTextDataXML(exampleXMLStream.str());
            if (!CifString::IsEmptyValue(exampleXMLStream.str()))
                _xsdWriter.WriteNewLine();
        }
        catch (const exception& exc)
        {
            cerr << "Skipping conversion to XML of the bad CIF "\
              "segment \"" << exampleCase[i] << "\"" << endl;
        }
    }

    _xsdWriter.WriteNewLine();

    _xsdWriter.Indent();
    _xsdWriter.WriteDocumentationClosingTag();
 
    _xsdWriter.DecrementIndent();

    _xsdWriter.Indent();
    _xsdWriter.WriteAnnotationClosingTag();
}


void PdbMlSchema::_WriteNonKeyItem(const string& itemName)
{
    string attribName;
    CifString::GetItemFromCifItem(attribName, itemName);

    _xsdWriter.Indent();
    _xsdWriter.WriteElementOpeningTag();

    _xsdWriter.WriteNameAttribute(attribName);

    _WriteItemAttributes(itemName, false);

    _xsdWriter.Indent();
    _xsdWriter.WriteElementClosingTag();
}


void PdbMlSchema::_WriteKeyItem(const string& itemName) 
{
    string attribName;
    CifString::GetItemFromCifItem(attribName, itemName);

    _xsdWriter.Indent();
    _xsdWriter.WriteAttributeOpeningTag();

    _xsdWriter.WriteNameAttribute(attribName);

    _WriteItemAttributes(itemName, true);

    _xsdWriter.Indent();
    _xsdWriter.WriteAttributeClosingTag();
}


void PdbMlSchema::_WriteCategoryKeys(const string& catName) 
{
    vector<string> keys = _dataInfo.GetCatKeys(catName);
    sort(keys.begin(), keys.end());

    if (keys.empty())
    {
        return;
    }

    // Write category keys
    vector<string> sortedCatKeys = keys;
    sort(sortedCatKeys.begin(), sortedCatKeys.end());

    _WriteComboKey(catName, sortedCatKeys, String::IntToString(0));

#ifdef VLAD_KEYS_ONLY_REFERENCES
    return;
#endif

    // Write other keys
    vector<vector<string> > parComboKeys;
    vector<vector<vector<vector<string> > > > allChildrenKeys;

    _FilterKeys(parComboKeys, allChildrenKeys, catName);

    // Id of 0 is reserved for all category keys

    unsigned keyId = 1;

    for (unsigned int keyI = 0; keyI < parComboKeys.size(); ++keyI)
    {
        if (parComboKeys[keyI].empty())
        {
            continue;
        }

        if (allChildrenKeys[keyI].empty())
        {
            continue; 
        }

#ifdef VLAD_KEYS_WITH_MANDATORY_DET_REFERENCES
        if (_AreAllKeyItems(parComboKeys[keyI]))
        {
            continue;
        }

        if (!_AreSupersetOfAllKeyItems(parComboKeys[keyI]))
        {
            continue;
        }
#endif

#ifdef VLAD_NO_KEYS_WITH_MANDATORY_DET_REFERENCES
        if (_AreAllKeyItems(parComboKeys[keyI]))
        {
            continue;
        }
#endif

#ifdef VLAD_NO_KEYS_NO_MANDATORY_DET_REFERENCES
        if (_AreAllKeyItems(parComboKeys[keyI]))
        {
            continue;
        }
#endif

        vector<string> sortedParComboKey = parComboKeys[keyI];
        sort(sortedParComboKey.begin(), sortedParComboKey.end());

        if (_WriteComboKey(catName, sortedParComboKey,
          String::IntToString(keyId), false))
        {
            keyId++;
        }
    }
}


void PdbMlSchema::_WriteCategoryKeysAndKeyrefs(const string& catName) 
{
    //const vector<string>& catKeys = _dataInfo.GetCatKeys(catName);

    vector<vector<string> > parComboKeys;
    vector<vector<vector<vector<string> > > > allChildrenKeys;

    _FilterKeys(parComboKeys, allChildrenKeys, catName);

    if (allChildrenKeys.empty())
        return;

    // Start from 1, as keyId of 0 is reserved for category primary key.
    // Category primary key consists of items that are all defined as
    // category keys.
    unsigned int keyId = 0;

    // This is used in order to prepend keyrefs that point to
    // category primary key.
    //unsigned int altKeyId = 0; 

    for (unsigned int keyI = 0; keyI < parComboKeys.size(); ++keyI)
    {
#ifdef VLAD_KEYS_ONLY_REFERENCES
        if (!_AreAllKeyItems(parComboKeys[keyI]))
        {
            continue;
        }
#endif

#ifdef VLAD_KEYS_WITH_MANDATORY_DET_REFERENCES
        if (!_AreSupersetOfAllKeyItems(parComboKeys[keyI]))
        {
            continue;
        }
#endif

        if (allChildrenKeys[keyI].empty())
            continue;

        const vector<vector<vector<string> > >& childrenKeys =
          allChildrenKeys[keyI];

        if (!_AreAllKeyItems(parComboKeys[keyI]))
        {
            keyId++;
        }

        for (unsigned int childI = 0; childI < childrenKeys.size(); ++childI)
        {
            string childCatName;
            if (!childrenKeys[childI].empty())
            {
                CifString::GetCategoryFromCifItem(childCatName,
                  childrenKeys[childI][0][0]); 
            }

            for (unsigned int childKeyI = 0; childKeyI <
              childrenKeys[childI].size(); ++childKeyI)
            {
                unsigned int usedKeyId = keyId;
                if (_AreAllKeyItems(parComboKeys[keyI]))
                {
                    usedKeyId = 0;
                }

                string keyRefPrefix = String::IntToString(keyI) + "_" +
                  String::IntToString(usedKeyId);

                string keyRefName = catName + "Keyref" + "_" +
                  keyRefPrefix + "_" +
                  String::IntToString(childI) + "_" +
                  String::IntToString(childKeyI);

                string keyName = _nsPrefix + catName + "Key" + "_" +
                  String::IntToString(usedKeyId);


                if (usedKeyId != 0)
                {
                    keyName = _nsPrefix + catName + "Unique" + "_" +
                      String::IntToString(usedKeyId);
                }

                string childCatElemName;
                PdbMlSchema::MakeCategoryElementName(childCatElemName,
                  childCatName);

                string xPath = _nsPrefix + childCatElemName + string("/") +
                  _nsPrefix + childCatName;

                multimap<string, string, StringLess> keyMap;

                for (unsigned int keyItemI = 0; keyItemI <
                  childrenKeys[childI][childKeyI].size(); ++keyItemI)
                {
                    string chAttribName;
                    CifString::GetItemFromCifItem(chAttribName,
                      childrenKeys[childI][childKeyI][keyItemI]);
                    multimap<string, string, StringLess>::value_type
                      valuePairIndex(parComboKeys[keyI][keyItemI],
                      chAttribName);
                    keyMap.insert(valuePairIndex);
                }

                vector<string> chKeys;
                for (multimap<string, string, StringLess>::iterator iter =
                  keyMap.begin(); iter != keyMap.end(); ++iter)
                {
                    chKeys.push_back((*iter).second);
                }

                _WriteKeyRef(keyRefName, keyName, xPath, chKeys, childCatName);
            } // for (all child's keys)
        } // for (all children)
    } // for (all parent combo keys)
}


void PdbMlSchema::_WriteItemAttributes(const string& itemName,
  const bool isKey)
{
    if (isKey)
    {
        _xsdWriter.WriteUseAttribute("required");
    }
    else
    {
        string minOccurs;

        if (_dataInfo.IsItemMandatory(itemName))
        {
            minOccurs = "1";
        }
        else
        {
            minOccurs = "0";
        }

        _xsdWriter.WriteMinOccursAttribute(minOccurs);
        _xsdWriter.WriteMaxOccursAttribute("1");

        if (_IsSkipParentItem(itemName) || _IsSkipChildItem(itemName))
        {
            _xsdWriter.WriteNillableAttribute("true");
        }
    }

    bool simpleTyping = _dataInfo.IsSimpleDataType(itemName);

    if (simpleTyping)
    {
        _WriteDataTypeAsAttribute(itemName);
    }

    _xsdWriter.WriteClosingBracket();

    _xsdWriter.IncrementIndent();

    _WriteItemDocumentation(itemName);

    if (!simpleTyping)
    {
        _WriteDataTypeAsElement(itemName);
    }

    _xsdWriter.DecrementIndent();
}


void PdbMlSchema::_WriteItemDocumentation(const string& itemName) 
{
    // Item level documentation  - 

    vector<string> description;
    description = _dataInfo.GetItemAttribute(itemName,
      CifString::CIF_DDL_CATEGORY_ITEM_DESCRIPTION,
      CifString::CIF_DDL_ITEM_DESCRIPTION);

    if (description.empty())
        description.push_back(string());
  
    string descriptionXML;
    format_cif_descriptionXML(descriptionXML, description[0]);

    _xsdWriter.Indent();
    _xsdWriter.WriteAnnotationOpeningTag();

    _xsdWriter.IncrementIndent();

    _xsdWriter.Indent();
    _xsdWriter.WriteDocumentationOpeningTag();
    _xsdWriter.WriteLangAttribute();
    _xsdWriter.WriteClosingBracket();

    if (!descriptionXML.empty())
    {
        _xsdWriter._FormatTextDataXML(descriptionXML);
        if (!CifString::IsEmptyValue(descriptionXML))
            _xsdWriter.WriteNewLine();
        if (!description[0].empty())
        {
            description[0].clear();
        }
    }

    const vector<string>& exampleCase = _dataInfo.GetItemAttribute(itemName,
      CifString::CIF_DDL_CATEGORY_ITEM_EXAMPLES,
      CifString::CIF_DDL_ITEM_CASE);

    const vector<string>& exampleDetail = _dataInfo.GetItemAttribute(itemName,
      CifString::CIF_DDL_CATEGORY_ITEM_EXAMPLES,
      CifString::CIF_DDL_ITEM_DETAIL);

    for (unsigned int i=0; i < exampleCase.size(); i++)
    {
        if (!exampleDetail.empty() && exampleDetail.size() > i)
        {
            _xsdWriter._FormatTextDataXML(exampleDetail[i]);
            if (!CifString::IsEmptyValue(exampleDetail[i]))
                _xsdWriter.WriteNewLine();
        }
        _xsdWriter._FormatTextDataXML(exampleCase[i]);
        if (!CifString::IsEmptyValue(exampleCase[i]))
            _xsdWriter.WriteNewLine();
    }

    _xsdWriter.Indent();
    _xsdWriter.WriteDocumentationClosingTag();
  
    _xsdWriter.DecrementIndent();

    _xsdWriter.Indent();
    _xsdWriter.WriteAnnotationClosingTag();
}


void PdbMlSchema::_WriteDataTypeAsElement(const string& itemName) 
{

  const vector<string>& rangeMin = _dataInfo.GetItemAttribute(itemName,
    CifString::CIF_DDL_CATEGORY_ITEM_RANGE,
    CifString::CIF_DDL_ITEM_MINIMUM);

  const vector<string>& rangeMax = _dataInfo.GetItemAttribute(itemName,
    CifString::CIF_DDL_CATEGORY_ITEM_RANGE,
    CifString::CIF_DDL_ITEM_MAXIMUM);

  bool iRange = false;

  if ((!rangeMin.empty()) && (!rangeMax.empty()) &&
    rangeMin.size() == rangeMax.size() && rangeMax.size() > 0)
  {
      iRange = true;
  }

  // 
  //  Assume that we may either have  range restrictions or an enumeration
  //  but not both!  Treat these cases first...
  //

  const vector<string>& enums = _dataInfo.GetItemAttribute(itemName,
    CifString::CIF_DDL_CATEGORY_ITEM_ENUMERATION,
    CifString::CIF_DDL_ITEM_VALUE);

  const vector<string>& units = _dataInfo.GetItemAttribute(itemName,
    CifString::CIF_DDL_CATEGORY_ITEM_UNITS,
    CifString::CIF_DDL_ITEM_CODE);

  eTypeCode iType = _dataInfo._GetDataType(itemName);

  if (!enums.empty())
  {
    _xsdWriter.Indent();
    _xsdWriter.WriteSimpleTypeOpeningTag();

    _xsdWriter.IncrementIndent();

    _xsdWriter.Indent();
    _xsdWriter.WriteRestrictionOpeningTag();
    _xsdWriter.WriteBaseAttribute(iType);
    _xsdWriter.WriteClosingBracket();

    _xsdWriter.IncrementIndent();
    for (unsigned int i=0; i < enums.size(); i++)
    {
        _xsdWriter.Indent();
        _xsdWriter.WriteEnumerationOpeningTag();
        _xsdWriter.WriteValueAttribute(enums[i], true);
        _xsdWriter.WriteClosingTag();
    }
    _xsdWriter.DecrementIndent();

    _xsdWriter.Indent();
    _xsdWriter.WriteRestrictionClosingTag();

    _xsdWriter.DecrementIndent();

    _xsdWriter.Indent();
    _xsdWriter.WriteSimpleTypeClosingTag();
  }
  else if (iRange)
  {
    _xsdWriter.Indent();
    _xsdWriter.WriteSimpleTypeOpeningTag();

    _xsdWriter.IncrementIndent();

    _xsdWriter.Indent();
    _xsdWriter.WriteUnionOpeningTag();

    _xsdWriter.IncrementIndent();

    for (unsigned int i=0; i < rangeMin.size(); i++)
    {
      _xsdWriter.Indent();
      _xsdWriter.WriteSimpleTypeOpeningTag();

      _xsdWriter.IncrementIndent();

      _xsdWriter.Indent();
      _xsdWriter.WriteRestrictionOpeningTag();
      _xsdWriter.WriteBaseAttribute(iType);
      _xsdWriter.WriteClosingBracket();

      _xsdWriter.IncrementIndent();
      if (String::IsCiEqual(rangeMin[i], rangeMax[i]))
      {

        if (!CifString::IsEmptyValue(rangeMin[i]))
        {
          _xsdWriter.Indent();
          _xsdWriter.WriteMinInclusiveOpeningTag();
          _xsdWriter.WriteValueAttribute(rangeMin[i]);
          _xsdWriter.WriteClosingTag();
	}

	if (!CifString::IsEmptyValue(rangeMax[i]))
        {
          _xsdWriter.Indent();
          _xsdWriter.WriteMaxInclusiveOpeningTag();
          _xsdWriter.WriteValueAttribute(rangeMax[i]);
          _xsdWriter.WriteClosingTag();
	}
      }
      else
      {
	if (!CifString::IsEmptyValue(rangeMin[i]))
        {
          _xsdWriter.Indent();
          _xsdWriter.WriteMinExclusiveOpeningTag();
          _xsdWriter.WriteValueAttribute(rangeMin[i]);
          _xsdWriter.WriteClosingTag();
	}
	if (!CifString::IsEmptyValue(rangeMax[i]))
        {
          _xsdWriter.Indent();
          _xsdWriter.WriteMaxExclusiveOpeningTag();
          _xsdWriter.WriteValueAttribute(rangeMax[i]);
          _xsdWriter.WriteClosingTag();
	}
      }
      _xsdWriter.DecrementIndent();

      _xsdWriter.Indent();
      _xsdWriter.WriteRestrictionClosingTag();

      _xsdWriter.DecrementIndent();

      _xsdWriter.Indent();
      _xsdWriter.WriteSimpleTypeClosingTag();
    }

    _xsdWriter.DecrementIndent();

    _xsdWriter.Indent();
    _xsdWriter.WriteUnionClosingTag();

    _xsdWriter.DecrementIndent();

    _xsdWriter.Indent();
    _xsdWriter.WriteSimpleTypeClosingTag();
  }
  else if (!units.empty())
  {
      _xsdWriter.Indent();
      _xsdWriter.WriteComplexTypeOpeningTag(true);

      _xsdWriter.IncrementIndent();

      _xsdWriter.Indent();
      _xsdWriter.WriteSimpleContentOpeningTag();

      _xsdWriter.IncrementIndent();

      _xsdWriter.Indent();
      _xsdWriter.WriteExtensionOpeningTag();
      _xsdWriter.WriteBaseAttribute(iType);
      _xsdWriter.WriteClosingBracket();

      _xsdWriter.IncrementIndent();

      _xsdWriter.Indent();
      _xsdWriter.WriteAttributeOpeningTag();
      _xsdWriter.WriteFixedAttribute(units[0]);
      _xsdWriter.WriteNameAttribute("units");
      _xsdWriter._WriteDataTypeAsAttribute(eTYPE_CODE_STRING);
      _xsdWriter.WriteUseAttribute("optional");
      _xsdWriter.WriteClosingTag();

      _xsdWriter.DecrementIndent();

      _xsdWriter.Indent();
      _xsdWriter.WriteExtensionClosingTag();

      _xsdWriter.DecrementIndent();

      _xsdWriter.Indent();
      _xsdWriter.WriteSimpleContentClosingTag();

      _xsdWriter.DecrementIndent();

      _xsdWriter.Indent();
      _xsdWriter.WriteComplexTypeClosingTag();
  }
}


void PdbMlSchema::_WriteDataTypeAsAttribute(const string& itemName)
{
    eTypeCode iType = _dataInfo._GetDataType(itemName);

    _xsdWriter._WriteDataTypeAsAttribute(iType);
}


void format_cif_example(string& obuf, const string& str)
{
    obuf.clear();

    if (str.empty())
    {
        return;
    }

    obuf += '\n';

    string tlineBuf;
    string::size_type currPos = 0;
 
    GetLineFromMultiline(tlineBuf, currPos, str);
    while (!tlineBuf.empty())
    {
        String::StripLeadingWs(tlineBuf);
        obuf += tlineBuf;
        GetLineFromMultiline(tlineBuf, currPos, str);
    }

    obuf += '\n';
    obuf += '\n';
}


void format_cif_descriptionXML(string& obuf, const string& str)
{
    // Read mmCIF description data and expand any explict item name reference
    // to attribute xxxx of category yyyy.
    obuf.clear();

    if (str.empty())
    {
        return;
    }

    string::size_type currPos = 0;
    string ilineBuf;

    GetLineFromMultiline(ilineBuf, currPos, str);
    while (!ilineBuf.empty())
    {
        String::StripLeadingWs(ilineBuf);

        unsigned int iFlag = 0;
        string olineBuf;
 
        string::size_type currLinePos = 0;

        string word;
        GetWordInLine(word, currLinePos, ilineBuf);

        while (!word.empty())
        {
            try
            {
                string categoryName;
                string attribName; 

                CifString::GetCategoryFromCifItem(categoryName, word); 
                CifString::GetItemFromCifItem(attribName, word); 

                if (attribName[attribName.size() - 1] == '.' ||
                  attribName[attribName.size() - 1] == ';' ||
                  attribName[attribName.size() - 1] == ',')
                {
                    attribName.erase(attribName.size() - 1);
                }

                iFlag++;

                olineBuf += "attribute ";
                olineBuf += attribName;
                olineBuf += " in category ";
                olineBuf += categoryName; 

                if (word[word.size() - 1] == '.' ||
                  word[word.size() - 1] == ';' ||
                  word[word.size() - 1] == ',')
                {
                    olineBuf += word[word.size() - 1];
                }
            }
            catch (EmptyValueException)
            {
                olineBuf += word;
            }

            olineBuf += " ";

            GetWordInLine(word, currLinePos, ilineBuf);
        }

        if (ilineBuf.empty())
        {
            obuf += '\n';
        }
        else
        {
            if (iFlag == 1)
            {
                // If only one CIF item in the line
                obuf += olineBuf;
            }
            else
            {
#ifdef VLAD_TEST
                if (iFlag != 0)
                    cerr << "Entered iFlag diff than 1 and 0. iFlag is: " <<
                      iFlag << " , and ilineBuf is \"" << ilineBuf << "\"" << endl;
#endif
                obuf += ilineBuf;
            }
        }

        GetLineFromMultiline(ilineBuf, currPos, str);
    }
}


void GetLineFromMultiline(string& line, string::size_type& currPos,
  const string& multiLine)
{
    if (currPos == string::npos)
    {
        line.clear();

        return;
    }

    string::size_type newLineInd = multiLine.find('\n', currPos);

    if (newLineInd != string::npos)
    {
        // Found a newline. Copy including the newline.
        line.assign(multiLine, currPos, newLineInd - currPos + 1);

        currPos = newLineInd + 1;
    }
    else
    {
        // Copy until the end. Do not insert newline.
        line.assign(multiLine, currPos, multiLine.size() - currPos);

        currPos = string::npos;
    }
}


void GetWordInLine(string& word, string::size_type& currLinePos,
  const string& line)
{
    if (currLinePos == string::npos)
    {
        word.clear();

        return;
    }

    string delimiters("\t ");

    // Find first non-delimiting char
    string::size_type nonDelimInd = line.find_first_not_of(delimiters,
      currLinePos);

    if (nonDelimInd == string::npos)
    {
        word.clear();

        currLinePos = string::npos;

        return;
    }

    // Find first delimiting char
    string::size_type delimInd = line.find_first_of(delimiters, nonDelimInd);

    if (delimInd != string::npos)
    {
        // Found a delimiter. Copy without delimiter.
        word.assign(line, nonDelimInd, delimInd - nonDelimInd);

        currLinePos = delimInd + 1;
    }
    else
    {
        // Copy until the end.
        word.assign(line, nonDelimInd, line.size() - nonDelimInd);

        currLinePos = string::npos;
    }
}


void PdbMlSchema::MakeCategoryElementName(string& catElemName,
  const string& catName)
{
    catElemName = catName + CATEGORY_ELEMENT_SUFFIX;
}


void PdbMlSchema::MakeCategoryTypeName(string& catTypeName,
  const string& catName)
{
    catTypeName = catName + CATEGORY_TYPE_SUFFIX;
}


bool PdbMlSchema::_WriteComboKey(const string& catName, 
  const vector<string>& keyItems, const string& append, const bool asXsdKey)
{
    bool emptyKeyItems = true;

    for (unsigned int i = 0; i < keyItems.size(); ++i)
    {
        if (_dataInfo.IsItemDefined(keyItems[i]))
        {
            emptyKeyItems = false;
            break;
        }
    }

    if (emptyKeyItems)
    {
        return (false);
    }

    string keyType;

    _xsdWriter.Indent();

    if (asXsdKey)
    {
        keyType = "Key";
        _xsdWriter.WriteKeyOpeningTag();
    }
    else
    {
        keyType = "Unique";
        _xsdWriter.WriteUniqueOpeningTag();
    }

    _xsdWriter.WriteNameAttribute(catName + keyType + "_" + append);
    _xsdWriter.WriteClosingBracket();

    _xsdWriter.IncrementIndent();

    _xsdWriter.Indent();
    _xsdWriter.WriteSelectorOpeningTag();

    string catElemName;
    PdbMlSchema::MakeCategoryElementName(catElemName, catName);
    _xsdWriter.WriteXpathAttribute(_nsPrefix + catElemName + string("/") +
      _nsPrefix + catName);
    _xsdWriter.WriteClosingTag();

    for (unsigned int i = 0; i < keyItems.size(); ++i)
    {
        if (!_dataInfo.IsItemDefined(keyItems[i]))
        {
            // VLAD - Log the undefined/invalid key item
            continue;
        }

        string attribName;
        CifString::GetItemFromCifItem(attribName, keyItems[i]);

        _xsdWriter.Indent();
        _xsdWriter.WriteFieldOpeningTag();

        if (_dataInfo.IsKeyItem(catName, attribName))
        {
            _xsdWriter.WriteXpathAttribute(attribName, true);
        }
        else
        {
            _xsdWriter.WriteXpathAttribute(_nsPrefix + attribName);
        }

        _xsdWriter.WriteClosingTag();
    }

    _xsdWriter.DecrementIndent();

    _xsdWriter.Indent();

    if (asXsdKey)
    {
        _xsdWriter.WriteKeyClosingTag();
    }
    else
    {
        _xsdWriter.WriteUniqueClosingTag();
    }

    return (true);
}


void PdbMlSchema::_WriteKeyRef(const string& keyRefName, const string& keyName,
  const string& xPath, const vector<string>& childKeys,
  const string& childCatName)
{
    if (childKeys.empty())
    {
        throw EmptyContainerException("No child keys for keyref \"" +
          keyRefName + "\"", "PdbMlSchema::_WriteKeyRef()");
    }

    _xsdWriter.Indent();
    _xsdWriter.WriteKeyrefOpeningTag();
    _xsdWriter.WriteNameAttribute(keyRefName);

    _xsdWriter.WriteReferAttribute(keyName);
    _xsdWriter.WriteClosingBracket();

    _xsdWriter.IncrementIndent();

    _xsdWriter.Indent();
    _xsdWriter.WriteSelectorOpeningTag();
    _xsdWriter.WriteXpathAttribute(xPath);
    _xsdWriter.WriteClosingTag();

    for (unsigned int keyI = 0; keyI < childKeys.size(); ++keyI)
    {
        _xsdWriter.Indent();
        _xsdWriter.WriteFieldOpeningTag();

        if (_dataInfo.IsKeyItem(childCatName, childKeys[keyI]))
        {
            _xsdWriter.WriteXpathAttribute(childKeys[keyI], true);
        }
        else
        {
            _xsdWriter.WriteXpathAttribute(_nsPrefix + childKeys[keyI]);
        }

        _xsdWriter.WriteClosingTag();
    }

    _xsdWriter.DecrementIndent();

    _xsdWriter.Indent();
    _xsdWriter.WriteKeyrefClosingTag();
}


void PdbMlSchema::_FilterKeys(vector<vector<string> >& parComboKeys,
  vector<vector<vector<vector<string> > > >& allChildrenKeys,
  const string& catName)
{
    //const vector<string>& catKeys = _dataInfo.GetCatKeys(catName);

    // Get all combo keys participating in parent-child relationships.
    const vector<vector<string> >& origParComboKeys =
     _parentChild.GetComboKeys(catName);

    for (unsigned int keyI = 0; keyI < origParComboKeys.size(); ++keyI)
    {
        const vector<string>& currOrigParComboKey = origParComboKeys[keyI];

        set<unsigned int> nonMandInd;

        _FindToSkipParentItemsIndices(nonMandInd, currOrigParComboKey);

#ifdef VLAD_KEYS_ONLY_REFERENCES
        // If this code is eliminated, the reduced parent key may not be
        // unique any more and XSD unique validation will fail. This code
        // can only be safely removed, once all the parent keys are supersets
        // of category key.
        if (!nonMandInd.empty())
            continue;
#endif

#ifdef VLAD_KEYS_WITH_MANDATORY_DET_REFERENCES
        // If this code is eliminated, the reduced parent key may not be
        // unique any more and XSD unique validation will fail. This code
        // can only be safely removed, once all the parent keys are supersets
        // of category key.
        if (!nonMandInd.empty())
            continue;
#endif

#ifdef VLAD_NO_KEYS_WITH_MANDATORY_DET_REFERENCES
        // If this code is eliminated, the reduced parent key may not be
        // unique any more and XSD unique validation will fail. This code
        // can only be safely removed, once all the parent keys are supersets
        // of category key.
        if (!nonMandInd.empty())
            continue;
#endif

#ifdef VLAD_NO_KEYS_NO_MANDATORY_DET_REFERENCES
        // If this code is eliminated, the reduced parent key may not be
        // unique any more and XSD unique validation will fail. This code
        // can only be safely removed, once all the parent keys are supersets
        // of category key.
        if (!nonMandInd.empty())
            continue;
#endif

        vector<vector<vector<string> > >& origChildrenKeys =
          _parentChild.GetChildrenKeys(currOrigParComboKey);

        // Pair of child index, child key index is a map key. Set of
        // to be skipped indices is map value.
        map<pair<unsigned int, unsigned int>, set<unsigned int> > chSkipInd;

        for (unsigned int childI = 0; childI < origChildrenKeys.size();
          ++childI)
        {
            for (unsigned int childKeyI = 0; childKeyI <
              origChildrenKeys[childI].size(); ++childKeyI)
            {
                const vector<string>& currChKey =
                  origChildrenKeys[childI][childKeyI];

                set<unsigned int> chKeySkipInd;

                _FindToSkipChildItemsIndices(chKeySkipInd, currChKey);

                // Insert a pair of (childI, childKeyI) and chKeySkipInd
                // in the map
                map<pair<unsigned int, unsigned int>, set<unsigned int> >::value_type valuePairIndex(make_pair(childI, childKeyI), chKeySkipInd);
                chSkipInd.insert(valuePairIndex);
            }
        }

        for (unsigned int parKeyI = 0; parKeyI < currOrigParComboKey.size();
          ++parKeyI)
        {
            unsigned int totalNumKeys = 0;
            unsigned int numIndexInKeys = 0;

            for (unsigned int childI = 0; childI < origChildrenKeys.size();
              ++childI)
            {
                for (unsigned int childKeyI = 0; childKeyI <
                  origChildrenKeys[childI].size(); ++childKeyI)
                {
                    totalNumKeys += origChildrenKeys[childI].size();

                    const set<unsigned int>& chSet =
                      chSkipInd[make_pair(childI, childKeyI)];

                    if (chSet.find(parKeyI) != chSet.end())
                    {
                        numIndexInKeys++;
                    }
                }
            }

            // If counter equals the sum of all keys, it means that that
            // item index is skipped in all keys.
            if (numIndexInKeys == totalNumKeys)
            {
                nonMandInd.insert(parKeyI);

                for (unsigned int childI = 0; childI < origChildrenKeys.size();
                  ++childI)
                {
                    for (unsigned int childKeyI = 0; childKeyI <
                      origChildrenKeys[childI].size(); ++childKeyI)
                    {
                        set<unsigned int>& chSet =
                          chSkipInd[make_pair(childI, childKeyI)];

                        chSet.erase(parKeyI);
                    }
                }
            }
        }

        vector<string> newParComboKey = currOrigParComboKey;
        _RemoveNonMandItems(newParComboKey, nonMandInd);
 
        parComboKeys.push_back(newParComboKey);
 
        vector<vector<vector<string> > > childrenKeys;

        for (unsigned int childI = 0; childI < origChildrenKeys.size();
          ++childI)
        {
            // Add empty vector
            vector<vector<string> > newChKeys;

            for (unsigned int childKeyI = 0; childKeyI <
              origChildrenKeys[childI].size(); ++childKeyI)
            {
                const vector<string>& currChKey =
                  origChildrenKeys[childI][childKeyI];

                vector<string> newChKey = currChKey;
                _RemoveNonMandItems(newChKey, nonMandInd);
 
                const set<unsigned int>& chSet =
                  chSkipInd[make_pair(childI, childKeyI)];

                if (!chSet.empty())
                {
                    continue;
                }

                if (!newChKey.empty())
                {
                    newChKeys.push_back(newChKey);
                }
            }

            if (!newChKeys.empty())
            {
                childrenKeys.push_back(newChKeys);
            }
        } // for (all child categories)

        allChildrenKeys.push_back(childrenKeys);
    } // for (all original parent combo keys)
}


bool PdbMlSchema::_AreAllKeyItems(const vector<string>& itemsNames)
{
    if (itemsNames.empty())
    {
        return (false);
    }

    string catName;
    CifString::GetCategoryFromCifItem(catName, itemsNames[0]);

    const vector<string>& catKeys = _dataInfo.GetCatKeys(catName);

    if (itemsNames.size() != catKeys.size())
    {
        return (false);
    }

    unsigned int keyItemsNum = 0;

    for (unsigned int i = 0; i < itemsNames.size(); ++i)
    {
        string attribName;
        CifString::GetItemFromCifItem(attribName, itemsNames[i]);
        if (_dataInfo.IsKeyItem(catName, attribName))
        {
            keyItemsNum++;
        }
    }

    if (keyItemsNum == catKeys.size())
    {
        return (true);
    }

    return (false);
}


bool PdbMlSchema::_AreSupersetOfAllKeyItems(const vector<string>& itemsNames)
{
    if (itemsNames.empty())
    {
        return (false);
    }

    string catName;
    CifString::GetCategoryFromCifItem(catName, itemsNames[0]);

    unsigned int keyItemsNum = 0;

    for (unsigned int i = 0; i < itemsNames.size(); ++i)
    {
        string attribName;
        CifString::GetItemFromCifItem(attribName, itemsNames[i]);
        if (_dataInfo.IsKeyItem(catName, attribName))
        {
            keyItemsNum++;
        }
    }

    const vector<string>& catKeys = _dataInfo.GetCatKeys(catName);

    if (keyItemsNum == catKeys.size())
    {
        return (true);
    }

    return (false);
}


void PdbMlSchema::_FindNonKeyItemsIndices(set<unsigned int>& nonMandIndices,
  const vector<string>& itemsNames)
{
    for (unsigned int indI = 0; indI < itemsNames.size(); ++indI)
    {
        string attribName;
        CifString::GetItemFromCifItem(attribName, itemsNames[indI]);

        string catName;
        CifString::GetCategoryFromCifItem(catName, itemsNames[indI]);

        if (!_dataInfo.IsKeyItem(catName, attribName))
        {
            nonMandIndices.insert(indI);
        }
    }
}


void PdbMlSchema::_FindNonMandItemsIndices(set<unsigned int>& nonMandIndices,
  const vector<string>& itemsNames)
{
    for (unsigned int indI = 0; indI < itemsNames.size(); ++indI)
    {
        if (!_dataInfo.IsItemMandatory(itemsNames[indI]))
        {
            nonMandIndices.insert(indI);
        }
    }
}


void PdbMlSchema::_FindToSkipParentItemsIndices(set<unsigned int>& indices,
  const vector<string>& itemsNames)
{
    for (unsigned int indI = 0; indI < itemsNames.size(); ++indI)
    {
        if (_IsSkipParentItem(itemsNames[indI]))
        {
            indices.insert(indI);
        }
    }
}


void PdbMlSchema::_FindToSkipChildItemsIndices(set<unsigned int>& indices,
  const vector<string>& itemsNames)
{
    for (unsigned int indI = 0; indI < itemsNames.size(); ++indI)
    {
        if (_IsSkipChildItem(itemsNames[indI]))
        {
            indices.insert(indI);
        }
    }
}


void PdbMlSchema::_RemoveNonMandItems(vector<string>& itemsNames,
  set<unsigned int>& nonMandIndices)
{
    for (set<unsigned int>::const_reverse_iterator it =
      nonMandIndices.rbegin(); it != nonMandIndices.rend(); ++it)
    {
        itemsNames.erase(itemsNames.begin() + (*it));
    }
}


bool PdbMlSchema::_IsSkipParentItem(const string& itemName)
{
#ifdef VLAD_KEYS_ONLY_REFERENCES
    string attribName;
    CifString::GetItemFromCifItem(attribName, itemName);

    string catName;
    CifString::GetCategoryFromCifItem(catName, itemName);

    if (!_dataInfo.IsKeyItem(catName, attribName))
    {
        return (true);
    }

    return (false);
#endif

#ifdef VLAD_KEYS_WITH_MANDATORY_DET_REFERENCES
    string attribName;
    CifString::GetItemFromCifItem(attribName, itemName);

    string catName;
    CifString::GetCategoryFromCifItem(catName, itemName);

    if (!_dataInfo.IsKeyItem(catName, attribName))
    {
        if (!_dataInfo.IsItemMandatory(itemName))
        {
            return (true);
        }

        if (CifExcept::CanBeInapplicable(itemName))
        {
            return (true);
        }
    }

    return (false);
#endif

#ifdef VLAD_NO_KEYS_WITH_MANDATORY_DET_REFERENCES
    string attribName;
    CifString::GetItemFromCifItem(attribName, itemName);

    string catName;
    CifString::GetCategoryFromCifItem(catName, itemName);

    if (!_dataInfo.IsItemMandatory(itemName))
    {
        return (true);
    }

    if (CifExcept::CanBeInapplicable(itemName))
    {
        return (true);
    }

    //if (CifExcept::IsBadParentRelation(itemName))
    //{
    //    return (true);
    //}

    return (false);
#endif

#ifdef VLAD_NO_KEYS_NO_MANDATORY_DET_REFERENCES
    string attribName;
    CifString::GetItemFromCifItem(attribName, itemName);

    string catName;
    CifString::GetCategoryFromCifItem(catName, itemName);

    if (CifExcept::CanBeInapplicable(itemName))
    {
        return (true);
    }

    //if (CifExcept::IsBadParentRelation(itemName))
    //{
    //    return (true);
    //}

    return (false);
#endif

    // VLAD - INVESTIGATE THIS
    //if (itemName == "_pdbx_poly_seq_scheme.auth_seq_num")
    //    return (true);
    //else if (itemName == "_entity_poly_seq.num")
    //    return (true);
    //else if (itemName == "_pdbx_poly_seq_scheme.seq_id")
    //    return (true);
    //else if (itemName == "_atom_site.auth_comp_id")
    //    return (true);
    //else if (itemName == "_atom_site.pdbx_PDB_ins_code")
    //    return (true);
    //else if (itemName == "_pdbx_poly_seq_scheme.seq_id")
    //    return (true);
    //else if (itemName == "_pdbx_poly_seq_scheme.pdb_ins_code")
    //    return (true);
    //else if (itemName == "_atom_site.auth_seq_id")
    //    return (true);
    //else if (itemName == "_pdbx_poly_seq_scheme.mon_id")
    //    return (true);
    //else if (itemName == "_atom_site.label_seq_id")
    //    return (true);
    //else if (itemName == "_struct_biol.id")
    //    return (true);
}


bool PdbMlSchema::_IsSkipChildItem(const string& itemName)
{
#ifdef VLAD_KEYS_ONLY_REFERENCES
    string attribName;
    CifString::GetItemFromCifItem(attribName, itemName);

    string catName;
    CifString::GetCategoryFromCifItem(catName, itemName);

    if (!_dataInfo.IsKeyItem(catName, attribName))
    {
        return (true);
    }

    return (false);
#endif

#ifdef VLAD_KEYS_WITH_MANDATORY_DET_REFERENCES
    string attribName;
    CifString::GetItemFromCifItem(attribName, itemName);

    string catName;
    CifString::GetCategoryFromCifItem(catName, itemName);

    if (!_dataInfo.IsKeyItem(catName, attribName))
    {
        if (!_dataInfo.IsItemMandatory(itemName))
        {
            return (true);
        }

        if (CifExcept::CanBeInapplicable(itemName))
        {
            return (true);
        }
    }

    return (false);
#endif

#ifdef VLAD_NO_KEYS_WITH_MANDATORY_DET_REFERENCES
    string attribName;
    CifString::GetItemFromCifItem(attribName, itemName);

    string catName;
    CifString::GetCategoryFromCifItem(catName, itemName);

    if (!_dataInfo.IsItemMandatory(itemName))
    {
        return (true);
    }

    if (CifExcept::CanBeInapplicable(itemName))
    {
        return (true);
    }

    if (CifExcept::IsBadChildRelation(itemName))
    {
        return (true);
    }

    return (false);
#endif

#ifdef VLAD_NO_KEYS_NO_MANDATORY_DET_REFERENCES
    string attribName;
    CifString::GetItemFromCifItem(attribName, itemName);

    string catName;
    CifString::GetCategoryFromCifItem(catName, itemName);

    if (CifExcept::CanBeInapplicable(itemName))
    {
        return (true);
    }

    if (CifExcept::IsBadChildRelation(itemName))
    {
        return (true);
    }

    return (false);
#endif


    // VLAD - DOCUMENT: These items are child combo keys and can have
    // values of "?"

    //if (itemName == "_pdbx_refine_tls_group.beg_label_asym_id")
    //    return (true);
    //else if (itemName == "_software.citation_id")
    //    return (true);
}

